"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const utils_1 = require("@code-engine/utils");
const validate_1 = require("@code-engine/validate");
const workers_1 = require("@code-engine/workers");
const events_1 = require("events");
const ono_1 = require("ono");
const os = require("os");
const build_pipeline_1 = require("./build/build-pipeline");
const normalize_plugin_1 = require("./plugins/normalize-plugin");
const plugin_controller_1 = require("./plugins/plugin-controller");
const instances = [];
// Hacky workaround to inherit from EventEmitter while also implementing the CodeEngineEventEmitter interface
const codeEngineEventEmitter = events_1.EventEmitter;
/**
 * The main CodeEngine class.
 */
class CodeEngine extends codeEngineEventEmitter {
    constructor(config = {}) {
        super();
        /** @internal */
        this._isDisposed = false;
        let context = {
            cwd: config.cwd || process.cwd(),
            concurrency: validate_1.validate.number.integer.positive(config.concurrency, "concurrency", os.cpus().length),
            dev: config.dev === undefined ? process.env.NODE_ENV === "development" : config.dev,
            debug: config.debug === undefined ? Boolean(process.env.DEBUG) : config.debug,
            log: undefined,
        };
        utils_1.createLogEmitter(this, context);
        this._context = context;
        this._buildPipeline = new build_pipeline_1.BuildPipeline(this);
        this._workerPool = new workers_1.WorkerPool(this, context);
        instances.push(this);
    }
    /**
     * Returns all CodeEngine instances that have been created and not yet disposed.
     */
    static get instances() {
        return instances.slice();
    }
    /**
     * Indicates whether the `dispose()` method has been called.
     * Once disposed, a CodeEngine instance is no longer usable.
     */
    get isDisposed() {
        return this._isDisposed;
    }
    /**
     * Loads one or more CodeEngine plugins.
     */
    async use(...plugins) {
        this._assertNotDisposed();
        for (let pluginDefinition of plugins) {
            // Normalize the plugin
            let defaultName = `Plugin ${this._buildPipeline.size + 1}`;
            let plugin = await normalize_plugin_1.normalizePlugin(pluginDefinition, this._workerPool, defaultName);
            let controller = new plugin_controller_1.PluginController(plugin);
            // Register any event handlers
            controller.onBuildStarting && this.on("buildStarting" /* BuildStarting */, controller.onBuildStarting.bind(controller));
            controller.onBuildFinished && this.on("buildFinished" /* BuildFinished */, controller.onBuildFinished.bind(controller));
            controller.onFileChanged && this.on("fileChanged" /* FileChanged */, controller.onFileChanged.bind(controller));
            controller.onError && this.on("error" /* Error */, controller.onError.bind(controller));
            controller.onLog && this.on("log" /* Log */, controller.onLog.bind(controller));
            // Add the plugin to the build pipeline
            this._buildPipeline.add(controller);
        }
    }
    /**
     * Imports one or more JavaScript modules in all worker threads.
     */
    async import(...modules) {
        this._assertNotDisposed();
        for (let moduleDefinition of modules) {
            await this._workerPool.importModule(moduleDefinition);
        }
    }
    /**
     * Deletes the contents of the destination(s).
     */
    async clean() {
        this._assertNotDisposed();
        let context = { ...this._context };
        return this._buildPipeline.clean(context);
    }
    /**
     * Runs a full build of all source files.
     */
    async build() {
        this._assertNotDisposed();
        let context = { ...this._context };
        try {
            let summary = await this._buildPipeline.build(context);
            return summary;
        }
        catch (error) {
            // Emit the "Error" event if there are any handlers registered
            if (this.listenerCount("error" /* Error */) > 0) {
                this.emit("error" /* Error */, error, context);
            }
            // Re-throw the error, regardless of whether there were event handlers
            throw error;
        }
    }
    /**
     * Watches source files for changes and runs incremental re-builds whenever changes are detected.
     *
     * @param delay
     * The time (in milliseconds) to wait after a file change is detected before starting a build.
     * This allows multiple files that are changed together to all be re-built together.
     */
    watch(delay) {
        delay = validate_1.validate.number.integer.positive(delay, "watch delay", 300);
        this._assertNotDisposed();
        let context = { ...this._context };
        this._buildPipeline.watch(delay, context);
    }
    /**
     * Releases system resources that are held by this CodeEngine instance.
     * Once `dispose()` is called, the CodeEngine instance is no longer usable.
     */
    async dispose() {
        if (this._isDisposed) {
            return;
        }
        this._isDisposed = true;
        let index = instances.indexOf(this);
        index >= 0 && instances.splice(index, 1);
        let context = { ...this._context };
        await Promise.all([
            this._buildPipeline.dispose(context),
            this._workerPool.dispose(),
        ]);
    }
    /**
     * Disposes all CodeEngine instances.
     */
    // tslint:disable-next-line: member-ordering
    static async disposeAll() {
        await Promise.all(instances.splice(0, instances.length).map((engine) => engine.dispose()));
    }
    /**
     * Returns a string representation of the CodeEngine instance.
     */
    toString() {
        if (this._isDisposed) {
            return `CodeEngine (disposed)`;
        }
        else {
            return `CodeEngine (${this._buildPipeline.size} plugins)`;
        }
    }
    /**
     * Returns the name to use for `Object.toString()`.
     */
    get [Symbol.toStringTag]() {
        return "CodeEngine";
    }
    /**
     * Throws an error if the `CodeEngine` has been disposed.
     * @internal
     */
    _assertNotDisposed() {
        if (this.isDisposed) {
            throw ono_1.ono(`CodeEngine cannot be used after it has been disposed.`);
        }
    }
}
exports.CodeEngine = CodeEngine;
//# sourceMappingURL=code-engine.js.map